########################################################################
#
# smartie.py
#
#########################################################################

"""smartie: CCP4 logfile parsing functions

The smartie module provides a set of classes and methods for parsing
logfiles from CCP4i and CCP4 programs. The central class is the 'logfile',
which provides a basic DOM-like description of a logfile and its
contents. Other classes provide descriptions of smaller chunks of logfile
features (programs, tables, keytext data and CCP4i informational messages).

The name 'smartie' reflects the module's origins as the intended driver
for a 'smart logfile browser'.

Some additional documentation material is also available in the file
smartie_overview.html."""

__cvs_id__ = "$Id: smartie.py,v 1.1 2007/02/15 10:21:50 wy45 Exp $"
__version__ = "0.0.2"

#######################################################################
# Import modules that this module depends on
#######################################################################
import sys
import os
import re
import copy
import linecache

#######################################################################
# Class definitions
#######################################################################

# buffer
#
# A class to store sets of lines which can then
# be passed to regular expression matching functions
#
class buffer:
    """Buffer object for parsing log files"""
    # Initialise the buffer object
    # If greater than zero, maxsize is the maximum
    # number of lines that the buffer should hold
    def __init__(self,maxsize=0):
        self.__contents = []
        self.__maxsize = maxsize
        
    # Append a line to the buffer
    def append(self,line):
        if self.__maxsize > 0 and len(self.__contents) > self.__maxsize:
            self.__contents = self.__contents[1:]
        self.__contents.append(line.rstrip("\n"))
        
    # Count number of lines in the buffer
    def len(self):
        return len(self.__contents)
    
    # Return n'th line (without newline) indexed by line number
    def line(self,n):
        return self.__contents[n]
    
    # Return tail of the buffer (newline separated)
    def tail(self,n=10):
        # Default number of lines is 10
        ##print "buffer.tail: n="+str(n)
        nend = self.len()
        nstart = nend - n
        ##print "buffer.tail: nstart="+str(nstart)+" nend="+str(nend)
        if nstart < 0:
            nstart = 0
            ##print "buffer.tail: nstart adjusted, now="+str(nstart)
        return self.contents(nstart,nend)
    
    # Return set of lines as a list
    def getlines(self,n,m):
        return self.__contents[n:m]
        
    # Return the specified lines from the buffer (newline separated)
    def contents(self,n,m):
        str = ""
        subset = self.getlines(n,m)
        for line in subset:
            str = str + line + "\n"
        return str.rstrip("\n")
    
    # Clear the buffer
    def clear(self):
        self.__contents[0:self.len()] = []
        return

#
# tablebuffer
#
# Buffer class specialised for handling tables
# Subclasses buffer
class tablebuffer(buffer):
    """Buffer object specialised for dealing with CCP4 tables"""

    def __init__(self,maxsize=0):
        self.__hasTABLE = False
        self.__hasGRAPHS = False
        self.__ndoubledollar = 0
        buffer.__init__(self,maxsize)

    # Override the append to also check for
    # table features
    def append(self,line):
        # Check if line has "$TABLE"
        hasTABLE = re.compile(r"\$TABLE *:").search(line)
        if hasTABLE:
            self.__hasTABLE = True
            ##print "Found TABLE keyword"
        # Check if line has "$(GRAPHS|SCATTER)"
        if self.__hasTABLE:
            buffer.append(self,line)
            hasGRAPHS = re.compile(r"\$(GRAPHS|SCATTER)").search(line)
            if hasGRAPHS:
                self.__hasGRAPHS = True
                ##print "Found GRAPHS/SCATTER keyword"
            # Check if line has "$$"
            if self.__hasGRAPHS:
                ##hasdoubledollar = re.compile(r"\$\$")
                ndoubledollar = line.count("$$")
                if ndoubledollar > 0:
                    self.__ndoubledollar = self.__ndoubledollar + ndoubledollar
                    ##print "line: "+str(line)
                    ##print "Found $$ symbol ("+str(self.__ndoubledollar)+" occurances)"

    # Check if we have a complete table yet
    def complete(self):
        ##print "tablebuffer.complete: hasTABLE  = "+str(self.__hasTABLE)
        ##print "tablebuffer.complete: hasGRAPHS = "+str(self.__hasGRAPHS)
        ##print "tablebuffer.complete: n$$       = "+str(self.__ndoubledollar)
        if self.__hasTABLE and self.__hasGRAPHS and self.__ndoubledollar == 4:
            return True
        else:
            return False

    # Override the clear to also reset the
    # flags
    def clear(self):
        self.__hasTABLE = False
        self.__hasGRAPHS = False
        self.__ndoubledollar = 0
        buffer.clear(self)

# logfile
#
# Abstract description of a CCP4 logfile
#
class logfile:
    """Object describing for program logfile.

    A logfile object is populated and returned by the
    parselog() function. This takes a file name as a single
    compulsory argument; the optional 'progress' argument
    specifies a number of lines at which to report progress
    when parsing the file.

    A logfile object holds lists of 'programs', 'tables',
    'keytext messages' and 'CCP4i information messages',
    plus a master list of 'fragments' (which can be any of
    the above). There are methods to allow access to each of
    these lists."""
    
    # Initialise the logfile object
    def __init__(self,filename):
        # Source file
        self.__filename = filename
        # CCP4i header and tail
        self.__isccp4i = False
        self.__ccp4i_header = []
        self.__ccp4i_tail = []
        # List of fragments, programs, tables, keytexts
        # and ccp4i_info
        self.__fragments = []
        self.__programs = []
        self.__tables = []
        self.__keytexts = []
        self.__ccp4i_info = []

    def __nonzero__(self):
        return False

    def append_ccp4i_header(self,line):
        """Append a line of text to the CCP4i header."""
        # FIXME should be internally accessible only?
        self.__ccp4i_header.append(line)
        self.__isccp4i = True

    def ccp4i_header(self):
        """Return the CCP4i header content."""
        return self.__ccp4i_header

    def append_ccp4i_tail(self,line):
        """Append a line of text to the CCP4i tail."""
        # FIXME should be internally accessible only?
        self.__ccp4i_tail.append(line)
        self.__isccp4i = True

    def ccp4i_tail(self):
        """Return the CCP4i tail content."""
        return self.__ccp4i_tail

    def isccp4i(self):
        """Return True if the logfile appears to be from CCP4i."""
        return self.__isccp4i

    def filename(self):
        """Return the filename of the source logfile."""
        return self.__filename

    def newfragment(self):
        """Generate a new fragement and add to the logfile.

        Returns a new fragment object and calls addfragment to
        add it to the list of fragments for this logfile."""
        newfragment = fragment()
        self.addfragment(newfragment)
        return newfragment

    def addfragment(self,fragment):
        """Add an existing fragment-like object to the logfile."""
        self.__fragments.append(fragment)

    def nfragments(self):
        """Return the number of fragments."""
        return len(self.__fragments)

    def fragment(self,i):
        """Return the i'th fragment in the logfile.

        A fragment can be a program, table, CCP4i message or
        keytext object.
        Note that i counts up from zero."""
        try:
            return self.__fragments[i]
        except:
            raise

    def addprogram(self):
        """Add a new program object to the logfile."""
        # FIXME should be internally accessible only?
        newprogram = program()
        self.__programs.append(newprogram)
        self.addfragment(newprogram)
        return newprogram

    def nprograms(self):
        """Return the number of program objects."""
        return len(self.__programs)

    def program(self,i):
        """Return the i'th program object in the logfile.

        Note that i counts up starting from zero."""
        try:
            return self.__programs[i]
        except:
            raise

    def addtable(self,thistable=False,tabletext=""):
        """Add a table object to the list of tables.

        If an existing table object is specified with the
        thistable argument then this is appended to the
        list. Otherwise a new table object is created. In
        that case, if tabletext is supplied then this is
        used to populate the table object; otherwise the
        new table object is empty."""
        
        # FIXME should be internally accessible only?
        if thistable:
            # Table object supplied
            self.__tables.append(thistable)
            return thistable
        else:
            # Make a new table
            if tabletext:
                new_table = table(tabletext)
            else:
                new_table = table()
            self.__tables.append(new_table)
            return new_table

    def ntables(self):
        """Return the number of tables in the logfile."""
        return len(self.__tables)

    def table(self,i):
        """Return the i'th table object in the logfile.

        Note that i counts up starting from zero."""
        try:
            return self.__tables[i]
        except:
            raise

    def addkeytext(self,thiskeytext=False,name="",junk_text="",message=""):
        """Add a keytext object to the list of keytexts.

        If an existing keytext object is supplied with the
        thiskeytext argument this is appended to the list.
        Otherwise a new keytext object is created and
        populated with the contents of the name, junk_text
        and message arguments."""
        
        # FIXME should be internally accessible only?
        if thiskeytext:
            # Table object supplied
            self.__keytexts.append(thiskeytext)
            return thiskeytext
        else:
            # Make a new keytext
            new_keytext = keytext(name,junk_text,message)
            self.__keytexts.append(new_keytext)
            return new_keytext

    def nkeytexts(self):
        """Return the number of keytexts in the logfile."""
        return len(self.__keytexts)

    def keytext(self,i):
        """Return the i'th keytext object in the logfile.

        Note that i counts up starting from zero."""
        try:
            return self.__keytexts[i]
        except:
            raise

    def addccp4i_info(self):
        """Add another ccp4i_info object to the logfile.
        
        Creates a new ccp4i_info object is created and
        added to the list of fragments and list of
        CCP4i information messages found in the logfile."""
        
        # FIXME should be internally accessible only?
        # Make a new ccp4i_info object
        newccp4i_info = ccp4i_info()
        self.__ccp4i_info.append(newccp4i_info)
        self.addfragment(newccp4i_info)
        return newccp4i_info

    def nccp4i_info(self):
        """Return the number of ccp4i_info messages."""
        return len(self.__ccp4i_info)

    def ccp4i_info(self,i):
        """Return the i'th ccp4i_info object in the logfile.

        Note that i counts up starting from zero."""
        try:
            return self.__ccp4i_info[i]
        except:
            raise

    def set_fragment_start(self,line_no):
        """Set the start line of the most recent fragment.

        The most recent fragment is the last fragment object
        in the list of fragments for this logfile. 'line_no'
        is the current line number in the source file.

        If the fragment has an 'nlines' attribute then this
        is taken to be the offset from the current line back
        to the start of the fragment. If nlines is not
        present then the fragment is taken to start after the
        end of the previous fragment. If there is no previous
        fragment then it is assumed to start from the first
        line of the file"""
        fragment = self.__fragments[-1]
        fragment.set_attribute("source_file",self.__filename)
        if fragment.has_attribute("nlines"):
            # Calculate the start of the fragment from the
            # current position
            offset = fragment.nlines
            fragment.set_startline(line_no-offset)
        else:
            if self.nfragments() > 1:
                # Assume that the fragment started from the
                # end of the previous fragment
                last_fragment = self.__fragments[-2]
                fragment.set_startline(last_fragment.get_endline()+1)
            else:
                # This is the first fragment
                fragment.set_startline(1)

    def set_fragment_end(self,line_no):
        """Set the end line of the most recent fragment.

        The most recent fragment is the last fragment object
        in the list of fragments for this logfile. 'line_no'
        is the current line number in the source file.

        The supplied line number is always taken as the last
        line number of the fragment. This method will also
        check the start line number and will attempt to set
        it to a reasonable value if it is not set: either the
        first line after the end of the previous fragment,
        or the start of the file (if there is no previous
        fragment)."""
        fragment = self.__fragments[-1]
        if fragment.get_endline() > -1:
            # Don't reset the value if it's already set
            return
        fragment.set_attribute("source_file",self.__filename)
        fragment.set_endline(line_no)
        # Check if the start is also set
        if fragment.get_startline() < 1:
            if self.nfragments() > 1:
                # Assume that the fragment started from the
                # end of the previous fragment
                last_fragment = self.__fragments[-2]
                fragment.set_startline(last_fragment.get_endline()+1)
            else:
                # This is the first fragment
                fragment.set_startline(1)

# fragment
#
# Abstract description of a generic logfile fragment
#
class fragment:
    """Object describing a generic fragment of a logfile.

    The fragment object is intended to represent any
    'top-level' fragment of a logfile, for example a
    program logfile or some output from a script that
    appears inbetween program logs.

    The fragment class is used as the base class for
    the program and ccp4i_info classes."""

    # Initialise the fragment object
    def __init__(self):
        # Initialise a dictionary to store arbitrary
        # attributes taken from the program logfile
        self.__dict = {}
        # List of tables
        self.__tables = []
        # List of keytexts
        self.__keytexts = []
        # For fragment retrieval
        self.set_source_file("")
        self.set_startline(-1)
        self.set_endline(-1)
        # Flags
        self.__nonzero = False

    def __nonzero__(self):
        return self.__nonzero

    def isfragment(self):
        # Return True if this is a raw fragment
        return True

    def isprogram(self):
        # Return True if this is a program
        return False

    def isccp4i_info(self):
        # Return True if this is a CCP4i information fragment
        return False

    def __setitem__(self,key,value):
        """Implements the functionality for program[key] = value

        Wrapper for the set_attribute method."""
        self.set_attribute(key,value)

    def __getitem__(self,key):
        """Implements the functionality for value = program[key]

        Wrapper for the get_attribute method."""
        return self.get_attribute(key)

    def __getattr__(self,key):
        """Implements the functionality for value = program.key

        Wrapper for the get_attribute method."""
        return self.get_attribute(key)

    def get_attribute(self,key):
        """Return the value of a logfile attribute.

        The key is a string specifying a particular logfile
        attribute. If the attribute has been read from the file
        then its value is returned, otherwise a KeyError
        exception is raised."""
        try:
            return self.__dict[key]
        except KeyError:
            raise KeyError, "Unknown attribute '"+str(key)+"'"

    def has_attribute(self,key):
        """Check whether a logfile attribute has been set.

        The key is a string specifying a particular logfile
        attribute. If the attribute has been set then this
        method returns True, otherwise it returns False."""
        return self.__dict.has_key(key)

    def attributes(self):
        """Return a list of all the logfile attributes.

        The list contains all the attributes that have been
        set for the logfile."""
        return self.__dict.keys()

    def set_attribute(self,key,value):
        """Set the value of a logfile attribute.

        The key is a string specifying a particular logfile
        attribute which will be assigned the given value.
        If the attribute doesn't exist then it will be created,
        if it does then the current value will be overwritten
        by the new one."""
        self.__dict[key] = value
        self.__nonzero = True

    def set_attributes_from_dictionary(self,dict):
        """Set the values of multiple logfile attributes.

        For each key in dictionary 'dict', the value of a
        logfile attribute with the same name as the key will
        be assigned the same value as that of the key."""
        for key in dict:
            self.__dict[key] = dict[key]
        self.__nonzero = True

    def addtable(self,tabletext=""):
        # Add another table object
        if tabletext:
            newtable = table(tabletext)
        else:
            newtable = table()
        self.__tables.append(newtable)
        self.__nonzero = True
        return newtable
    
    def ntables(self):
        """Return the number of tables found in the logfile fragment."""
        return len(self.__tables)

    def table(self,i):
        """Return the i'th table object.

        For example: program.table(i) returns the i'th table
        object associated with the program object. The methods
        of the table class can then be used to drill down into
        the contents of the table.

        Use the ntables method to get the total number of table
        objects associated with the program/fragment."""
        try:
            return self.__tables[i]
        except:
            raise

    def addkeytext(self,name="",junk_text="",message=""):
        # Add another keytext object
        # FIXME should be internally accessible only?
        newkeytext = keytext(name,junk_text,message)
        self.__keytexts.append(newkeytext)
        self.__nonzero = True
        return newkeytext

    def nkeytexts(self):
        """Return the number of keytexts found in the logfile fragment.

        'Keytexts' are warnings and messages issued by the
        ccperror/CCPERR functions within programs; see the
        loggraph format documentation for more information, e.g.
        http://www.ccp4.ac.uk/dist/html/loggraphformat.html"""
        return len(self.__keytexts)

    def keytext(self,i):
        """Return the i'th keytext object.
        
        For example: program.keytext(i) returns the i'th keytext
        object associated with the program object. The methods
        of the keytext class can then be used to drill down into
        the contents of the message.

        Use the nkeytexts method to get the total number of
        keytext objects associated with the program/fragment."""
        try:
            return self.__keytexts[i]
        except:
            raise

    def set_startline(self,line_no):
        """Set the start line of the fragment in the source document."""
        self.set_attribute("startline",line_no)

    def get_startline(self):
        """Get the start line of the fragment in the source document."""
        return self.get_attribute("startline")

    def set_endline(self,line_no):
        """Set the end line of the fragment in the source document."""
        self.set_attribute("endline",line_no)

    def get_endline(self):
        """Get the end line of the fragment in the source document."""
        return self.get_attribute("endline")

    def set_source_file(self,source_file):
        """Set the source document for the fragment.

        The source document is specified as the name of the file that
        the fragment is part of."""
        self.set_attribute("source_file",source_file)

    def get_source_file(self):
        """Get the source document for the fragment."""
        return self.get_attribute("source_file")

    def retrieve(self):
        """Retrieve the text associated with the fragment.

        This uses the linecache module - beware of problems with
        consuming too much memory if the cache isn't cleared."""
        text = ""
        # Retrieve the information
        filen = self.get_source_file()
        start = self.get_startline()
        end = self.get_endline()
        # Check for consistency and validity of lines
        if start < 0 and end < 0 or end < start:
            return ""
        # Fetch from a file if possible
        if os.path.isfile(filen):
            try:
                for i in range(start,end+1):
                    text = text+str(linecache.getline(filen,i))
                    linecache.clearcache()
                return text
            except:
                print "Exception raised in retrieve method:"
                print "\tSource file  = \""+str(filen)+"\""
                print "\tStart line   = "+str(start)
                print "\tEnd line     = "+str(end)
                print "\tCurrent line = "+str(i)
            raise
        # Otherwise return nothing
        return ""

#
# program
#
# Abstract description of the logfile for a single program
#
class program(fragment):
    """Object describing the log for a single program.

    program objects are instantiated and populated by parselog
    as part of the parsing process. The program object is
    intended to describe a fragment of logfile that corresponds
    to the run of a particular program, although in practice
    other types of logfile features (for example, 'interstitial'
    fragments i.e. bits of output inbetween program logs) are
    also assigned to program objects in the current version of
    smartie.

    A program object holds various attributes describing the
    logfile fragment in question, as well as a list of tables
    and keytext messages. A program object may also hold CCP4i
    information messages, however normally it will not hold this
    at the same time as actual program data.

    The attributes associated with the program object can be
    accessed using either of the syntaxes 'program['attribute']'
    or 'program.attribute'.

    For programs using the standard CCP4 banners, the following
    attributes may be defined:

    name: the name of the program from the CCP4 banner, or
    equivalent.
    
    version: the program version; for CCP4 programs, this is the
    version found in the program banner. For programs that don't
    explicitly give their own version number this will be the same
    as the CCP4 library version.
    
    date: the date string found in the CCP4 banner; it is
    typically the last date that the source code file was
    committed to CVS. It is not the date that the program was
    run on - for that, see the 'rundate' and 'runtime' attributes.

    ccp4version: the CCP4 library version as it appears in the
    program banner. Typically this includes only the major and
    minor version numbers, but not the patch level.

    user: the user id that appears in the CCP4 banner at runtime.

    runtime: the time of day that the program run started at as
    reported in the program banner.

    rundate: the date that the program run started at as
    reported in the program banner.

    termination_name: the program name as reported in the
    CCP4 termination message at the tail of the program log.

    termination_message: the message text displayed in the
    CCP4 termination message.

    usertime: the value of the 'user time' given at
    termination.

    systemtime: the value of the 'system time' given at
    termination.

    elapsedtime: the value of the 'elapsed time' given at
    termination.

    Note that not all these attributes may be defined, for
    example if the program fragment is an incomplete CCP4 log
    file or if the program is not a CCP4 program. Use the
    'attributes' method to get a list of the defined
    attributes."""
    
    # Initialise the program object
    def __init__(self):
        # Initialise the base class
        fragment.__init__(self)
        # Initialise program-specific flags and
        # attributes
        self.__isccp4 = False
        self.__termination = False
        
    def isprogram(self):
        return True

    def isfragment(self):
        return False
    
    def set_isccp4(self,isccp4):
        """Set whether the logfile fragment is from a CCP4 program or not.

        This method sets the value of the isccp4 flag to True
        if the logfile fragment is determined to be from a CCP4
        program, and False if not. Use the 'isccp4' method to
        return the value of this flag."""
        # Possibly this should be internally accessible only?
        self.__isccp4 = isccp4

    def isccp4(self):
        """Check if the logfile fragment is from a CCP4 program.

        This returns True if the fragment of logfile appeared to
        be from a CCP4 program, and False otherwise."""
        return self.__isccp4
        
    def set_termination(self,termination):
        """Set whether the logfile has a termination message.

        This sets the value of the 'termination' flag to be
        True if a termination message was found, and False if
        not. Use the 'termination' method to return the value
        of this flag."""
        # FIXME should be internally accessible only?
        self.__termination = termination

    def termination(self):
        """Check if the logfile fragment ends with a valid termination.

        This returns True if the fragment appeared to finish with a
        recognised termination message, False otherwise.
        Program fragments that do not end with a termination
        message may have terminated prematurely due to an error."""
        return self.__termination

#
# table
#
# Abstract description of a CCP4 formatted logfile table
#
class table:
    """Object describing a CCP4 logfile table

    The table class represents the various components of a table
    as output in CCP4 program logfiles. These tables are formatted
    in a standard way that enables the data that they contain to
    be displayed by the (x|j)loggraph programs.

    For a description of the loggraph format see the loggraph
    format documentation, e.g.
    http://www.ccp4.ac.uk/dist/html/loggraphformat.html

    A table consists of a number of columns of data, and a
    number of graphs which are defined as being a subset of these
    columns. Within smartie the table_column class represents an
    individual column, and the table_graph class represents an
    individual graph.

    A table object can be populated when it is created, by
    supplying it with text containing a CCP4-formatted table
    (typically, a fragment of logfile text). Alternatively an
    'empty' table can be instantiated and then populated using
    the methods of the objects.

    The contents of the table can be output in the correct
    CCP4 format using the 'show' and 'jloggraph' methods."""
    
    # Initialise the table object
    def __init__(self,tabletext=""):
        """Create a new table object.

        If tabletext contains the text of an existing
        CCP4-formatted table then the table object will
        attempt to parse the table and populate itself using
        the supplied data. Otherwise the table will be 'empty'
        i.e. containing no data."""
        
        # Table attributes
        self.__title = ""
        self.__type = ""
        self.__graphs = ""
        self.__columns = ""
        self.__text = ""
        self.__data = ""
        # Derived data
        self.__graph_list = []
        self.__column_list = []
        # Indicate the the object has been populated
        self.__table_parse_error = False
        self.__nonzero = False
        # The "raw" table data from the log file
        if tabletext:
            self.__rawtable = tabletext
            self.__buildtable(tabletext)
        else:
            self.__rawtable = ""

    def __nonzero__(self):
        """Builtin: provides the True/False test.

        A table object is nonzero if data has been loaded
        into it either at instantiation or subsequently
        using any of its methods."""
        return self.__nonzero

    def __buildtable(self,tabletext):
        """Internal: populates the table object from an existing
        formatted table."""
        
        # Set up the table object by parsing the
        # the supplied text
        tabledata = isccp4table(tabletext)
        if not tabledata:
            # No match
            return
        # Populate the table object
        self.settitle(tabledata["title"])
        self.settype(tabledata["type"])
        self.setgraphs(tabledata["graphs"])
        self.setcolumns(tabledata["columns"])
        self.settext(tabledata["text"])
        self.setdata(tabledata["data"])
        self.__nonzero = True

    def __populate_columns(self):
        """Internal: populates the table_column objects."""
        
        # Parse the raw data and populate the table_column
        # objects for this table
        i = 0
        for item in self.__data.split():
            #print "Append column "+str(i)+" ("+self.table_column(i).title()+")"
            self.table_column(i).append(item)
            i += 1
            if i == self.ncolumns():
                i = 0
        # If there are enough items then i should be
        # zero at the end
        if i != 0:
            print "**** Error in __populate_columns! ****"
            self.__table_parse_error = True

    def parse_error(self):
        """Check if the supplied table was parsed correctly.

        If there was a problem parsing the raw table text (for
        example if the table was incomplete or misformatted) then
        parse_error() will return True, otherwise it will be
        False."""
        
        # Check the table_parse_error flag
        return self.__table_parse_error

    def setrawtable(self,rawtable):
        """Store the 'raw' table data in the table object.

        The raw table data is the original text (for example, the
        fragment of log file text) supplied to the object to
        populate itself from."""
        
        # Store the "raw" table data
        self.__rawtable = rawtable

    def rawtable(self):
        # Return the raw table
        return self.__rawtable

    def settitle(self,title):
        # Store the table title
        self.__title = title
        self.__nonzero = True

    def title(self):
        # Return the table title
        return self.__title

    def settype(self,type):
        # Store the table graph type (either GRAPHS or SCATTER)
        self.__type = type
        self.__nonzero = True

    def type(self):
        # Return the table graph type
        return self.__type

    def setgraphs(self,graphs):
        # Store the graph titles and descriptions ("raw" format)
        # The graph data is in the form:
        # :graph1 name:graphtype:column_list: :graph2 name:graphtype:column_list: ...
        self.__graphs = graphs
        ##print "Raw graph description = \""+graphs+"\""
        # Create table_graph objects
        rgraph = re.compile(r":([^:]+):([^:]+):([^:]+):")
        for graph in rgraph.findall(graphs):
            ##print "Graph descriptor = \""+str(graph)+"\""
            new_graph = self.addgraph(graph[0])
            new_graph.setscaling(graph[1])
            new_graph.setcolumns(graph[2])
        self.__nonzero = True

    def graphs(self):
        # Return the graph titles and descriptions ("raw" format)
        return self.__graphs

    def setcolumns(self,columns):
        # Store the column titles ("raw" format)
        # This is a list of white-space separated strings
        self.__columns = columns
        # Create table_column objects
        for col in columns.split():
            self.addcolumn(col)
        # Attempt to populate the column objects
        if self.__data:
            self.__populate_columns()
        self.__nonzero = True

    def columns(self):
        # Return the column titles ("raw" format)
        return self.__columns

    def settext(self,text):
        # Store the arbitrary text from the table
        self.__text = text
        self.__nonzero = True

    def text(self):
        # Return the arbitrary text from the table
        return self.__text

    def setdata(self,data):
        # Store the data from the table ("raw" format)
        # This is a list of whitespace separated data items
        self.__data = data
        # Attempt to populate the column objects
        if self.ncolumns() > 0:
            self.__populate_columns()
        self.__nonzero = True

    def data(self):
        # Return the data from the table ("raw" format)
        return self.__data

    def ncolumns(self):
        # Return the number of table_column objects
        return len(self.__column_list)

    def addcolumn(self,title=""):
        # Add a new table_column object to the table
        new_column = table_column()
        self.__column_list.append(new_column)
        if title:
            new_column.settitle(title)
        return new_column

    def table_column(self,i):
        # Return the i'th table_column object
        try:
            return self.__column_list[i]
        except:
            raise

    def col(self,name):
        """Return the data in the column identified by 'name'.

        This method returns the data in the table column
        identified by 'name' (for example, 'Rfree'), as a
        list of values. (This is a copy of the list of values
        in the table_column object representing the column.)

        If the named column isn't found then a LookupError
        exception is raised."""
        # Identify the column corresponding to the
        # supplied name and return a copy of the data
        for i in range(0,self.ncolumns()):
            if self.table_column(i).title() == name:
               return copy.copy(self.table_column(i).data())
        # Column not found
        raise LookupError, "Column called '"+str(name)+"' not found"

    def ngraphs(self):
        # Return the number of table_graph objects
        return len(self.__graph_list)

    def addgraph(self,title=""):
        # Add a new graph object to the table
        new_graph = table_graph()
        self.__graph_list.append(new_graph)
        if title:
            new_graph.settitle(title)
        return new_graph

    def table_graph(self,i):
        # Return the i'th table_graph object
        try:
            return self.__graph_list[i]
        except:
            raise

    def jloggraph(self,codebase="",width=400,height=300):
        """Return a jloggraph-formatted table.

        This method returns the text for CCP4-formatted table
        from this object which includes the HTML tags required
        for the jloggraph Java applet.

        The codebase argument should specify the full path for
        the JLogGraph.class and JLogCanvas.class files required
        to run the applet (typically this is $CCP4/bin/)."""
        
        # Wraps the show method
        jloggraph = "<applet width=\""+str(width)+ \
                    "\" height=\""+str(height)+ \
                    "\" code=\"JLogGraph.class\"\n"+ \
                    "codebase=\""+str(codebase)+ \
                    "\"><param name=\"table\" value=\"\n"
        jloggraph = jloggraph+self.show()
        jloggraph = jloggraph+"\"><b>For inline graphs use a Java browser</b></applet>"
        return jloggraph

    def show(self):
        """Return the text of a CCP4-formatted table.

        The show method generates the text of the table based
        on the data stored in the object, with the correct
        tags defining the columns and graphs and which should
        be viewable in (x)loggraph.

        To generate jloggraph-formatted tables use the
        jloggraph method."""
        
        # Preamble
        tabletext = "$TABLE: "+self.title()+":\n$GRAPHS\n"
        # Graph descriptions
        for i in range(0,self.ngraphs()):
            graph = self.table_graph(i)
            tabletext = tabletext+" :"+ \
                        graph.title()+":"+ \
                        graph.scaling()+":"
            for col in graph.columns():
                tabletext = tabletext+str(col+1)+","
            tabletext = tabletext.rstrip(",")
            tabletext = tabletext+":\n"
        tabletext = tabletext+"$$\n"
        # Columns and rows
        ncolumns = self.ncolumns()
        if ncolumns > 0:
            nrows = len(self.table_column(0))
        else:
            nrows = 0
        # Determine field widths for printing
        field_width = []
        for i in range(0,ncolumns):
            max_width = len(self.table_column(i).title())
            for item in self.table_column(i).data():
                if len(str(item)) > max_width:
                    max_width = len(str(item))
            if max_width >= len(self.table_column(i).title()):
                # Put in an extra space again
                max_width = max_width+1 
            field_width.append(max_width)
        # Column titles
        for i in range(0,ncolumns):
            title = self.table_column(i).title()
            while len(title) < field_width[i]:
                title = " "+title
            tabletext = tabletext+" "+title
        tabletext = tabletext+" $$"
        # Arbitrary text
        if self.text():
            tabletext = tabletext+self.text()
        tabletext = tabletext+" $$\n"
        # The columns of data
        for i in range(0,nrows):
            for j in range(0,ncolumns):
                item = self.table_column(j)[i]
                while len(str(item)) < field_width[j]:
                    item = " "+str(item)
                tabletext = tabletext+" "+str(item)
            tabletext = tabletext+"\n"
        # End of table
        tabletext = tabletext+"$$"
        return tabletext

#
# table_graph
#
# Abstract description of a graph in a CCP4 logfile table
#
class table_graph:
    """Object describing a graph in a CCP4 logfile table.

    Tables in logfiles can contain any number of 'graphs',
    which are represented within smartie by table_graph
    objects.
    
    A graph is defined by a title, a scaling type, and a
    collection of table_columns storing columns of data."""
    
    # Initialise the table_graph object
    def __init__(self,title="",scaling="",column_list=[]):
        """Create a new table_graph object.

        The 'title' argument is a string containing the title
        for the graph.
        'scaling' is a string describing how the graph should
        be displayed within the (x|j)loggraph program.
        'column_list' is a list of integers corresponding to
        the columns in the table that holds the graph. The first
        column in the list will form the 'x' axis of the graph
        when displayed, the others will be displayed on the
        'y' axis."""
        
        self.__title = title
        self.__column_list = column_list
        self.__scaling = scaling
        if self.__title:
            self.__nonzero = True
        else:
            self.__nonzero = False

    def __nonzero__(self):
        return self.__nonzero

    def settitle(self,title):
        """Store the title of the graph."""
        self.__title = title

    def title(self):
        """Return the title of the graph."""
        return self.__title

    def setscaling(self,scaling):
        """Store the scaling description.
        
        This is a string which should take one of three possible
        forms, and which is an instruction to the display
        program on how to scale the graph data for display.

        'A' is 'fully automatic' scaling (the display program
        determines the scaling itself for both axes).
        'N' (for 'nought') is automatic y coordinate scaling, where
        the lowest limit on the y axis is 0s.
        'XMIN|XMAXxYMIN|YMAX' (where XMIN, XMAX and YMIN, YMAX are
        numbers) specifies the exact limits of both axes."""
        
        self.__scaling = scaling

    def scaling(self):
        """Return the scaling description."""
        return self.__scaling

    def setcolumns(self,columns):
        """Set the table_columns associated with the graph.

        The columns are specified as a string of the form
        e.g. '1,2,4,5'. Note that the column numbers are adjusted
        downwards by 1 to map onto Python numbering (which starts
        at zero)."""
        
        self.__column_list = []
        for i in columns.split(","):
            if str(i).strip().isdigit():
                self.__column_list.append(int(i)-1)

    def columns(self):
        """Return the list of columns associated with the graph.

        This is a list of integers corresponding to the columns
        in the table."""
        return self.__column_list

#
# table_column
#
# Abstract description of a column in a CCP4 logfile table
#
class table_column:
    """Object describing a column in a CCP4i logfile table"""
    # Initialise the table_column object
    def __init__(self,title=""):
        self.__title = title
        self.__data = []
        if self.__title:
            self.__nonzero = True
        else:
            self.__nonzero = False

    def __nonzero__(self):
        return self.__nonzero

    def __len__(self):
        return len(self.__data)

    def __getitem__(self,key):
        return self.__data[key]

    def settitle(self,title):
        # Set the title of the column
        self.__title = title

    def title(self):
        # Return the title of the column
        return self.__title

    def append(self,value):
        # Add a data value to the column
        self.__data.append(value)

    def data(self):
        # Retun the list of data values
        return self.__data

#
# keytext
#
# Abstract description of a CCP4 formatted keytext message
#
class keytext:
    """Object describing a keytext message in a CCP4 logfile"""
    # Initialise the keytext object
    def __init__(self,name="",junk_text="",message=""):
        self.setname(name)
        self.setjunk_text(junk_text)
        self.setmessage(message)

    def setname(self,name):
        # Set the name attribute
        self.__name = str(name).strip()

    def name(self):
        # Return the name attribute
        return self.__name

    def setjunk_text(self,junk_text):
        # Set the junk_text attribute
        self.__junk_text = str(junk_text).strip()

    def junk_text(self):
        # Return the junk_text attribute
        return self.__junk_text

    def setmessage(self,message):
        # Set the message attribute
        self.__message = str(message).strip()

    def message(self):
        # Return the message attribue
        return self.__message

#
# ccp4i_info
#
# Abstract description of a CCP4i information message
#
class ccp4i_info(fragment):
    """Object describing a CCP4i information message in a CCP4 logfile.

    The ccp4i_info class has the following attributes:

    'message': the text of the CCP4i information message."""
    # Initialise the keytext object
    
    def __init__(self):
        # Initialise the base class
        fragment.__init__(self)
        # Initialise program-specific flags and
        # attributes
        self.set_attribute("message","")
        
    def isccp4i_info(self):
        return True

    def isfragment(self):
        return False

#######################################################################
# External Functions - regular expression matching
#######################################################################

def isccp4banner(text):
    """Regular expression match to CCP4 program banner.

    Given a block of text, attempts to match it against
    regular expressions for a CCP4 program banner.
    Returns False if the match fails, otherwise returns
    a dictionary object populated with attributes
    derived from the supplied text.

    See the isccp4banner_standard and isccp4banner_phaser
    functions for descriptions of the attributes that are
    extracted."""
    # Try standard CCP4 banner
    result = isccp4banner_standard(text)
    if not result:
        # Try Phaser-style CCP4 banner
        result = isccp4banner_phaser(text)
    return result

# Match CCP4 program termination
def isccp4termination(text):
    """Regular expression match to CCP4 program termination.

    Given a block of text, attempts to match it against
    regular expressions for a CCP4 program termination.
    Returns False if the match fails, otherwise returns
    a dictionary object populated with attributes
    derived from the supplied text.

    See the isccp4termination_standard and
    isccp4termination_phaser functions for descriptions of
    the attributes that are extracted."""
    # Try standard CCP4 termination
    result = isccp4termination_standard(text)
    if not result:
        # Try Phaser-style CCP4 termination
        result = isccp4termination_phaser(text)
    return result

# Match standard CCP4 program banner
def isccp4banner_standard(text):
    """Test if text matches a standard CCP4 program banner.

    If the match fails then return False; if it succeeds then
    return a dictionary with the following keys:

    name: the name of the program from the CCP4 banner.
    
    version: the program version; for CCP4 programs, this is the
    version found in the program banner. For programs that don't
    explicitly give their own version number this will be the same
    as the CCP4 library version.
    
    date: the date string found in the CCP4 banner; it is
    typically the last date that the source code file was
    committed to CVS. (It is not the date that the program was
    run on - for that, use 'rundate' and 'runtime').

    ccp4version: the CCP4 library version as it appears in the
    program banner. Typically this includes only the major and
    minor version numbers, but not the patch level.

    user: the user id that appears in the CCP4 banner at runtime.

    runtime: the time of day that the program run started at as
    reported in the program banner.

    rundate: the date that the program run started at as
    reported in the program banner."""
    #print "************* TEXT **************"
    #print text
    #print "*********************************"
    #
    # Banner looks like:
    #  ###############################################################
    #  ###############################################################
    #  ###############################################################
    #  ### CCP4 5.99: Refmac_5.2.0019    version 5.2.0019  : 04/08/05##
    #  ###############################################################
    #
    #  User: pjx  Run date: 25/10/2005 Run time: 15:19:23
    #
    rbanner = re.compile(r" #{63,63}\n #{63,63}\n #{63,63}\n ### CCP4 ([0-9.]+): ([A-Za-z0-9_.]+) *version ([0-9.]+) *: ([0-9 /]+)##\n #{63,63}\n User: ([^ ]+) *Run date: ([0-9 /]+) Run time: ([0-9:]+) ?")
    banner = rbanner.search(text)
    result = dict()
    if banner:
        result["banner_text"] = banner.group(0)
        result["ccp4version"] = banner.group(1)
        result["name"] = banner.group(2)
        result["version"] = banner.group(3)
        result["date"] = banner.group(4)
        result["user"] = banner.group(5)
        result["rundate"] = banner.group(6)
        result["runtime"] = banner.group(7)
        result["nlines"] = banner.group(0).count("\n")
    return result

# Match standard CCP4 program termination
def isccp4termination_standard(text):
    """Test if text matches a standard CCP4 program termination.

    If the match fails then return False; if it succeeds then
    return a dictionary with the following keys:

    termination_name: the program name as reported in the
    CCP4 termination message at the tail of the program log.

    termination_message: the message text displayed in the
    CCP4 termination message, e.g. 'Normal termination'.

    usertime: the value of the 'user time' given at
    termination.

    systemtime: the value of the 'system time' given at
    termination.

    elapsedtime: the value of the 'elapsed time' given at
    termination."""
    #print "************* TEXT **************"
    #print text
    #print "*********************************"
    #
    # Termination looks like:
    #  Refmac_5.2.0019:  End of Refmac_5.2.0019
    # Times: User:       6.0s System:    0.4s Elapsed:     0:07
    #
    rterm = re.compile(r" *([A-Za-z0-9_]+): *([^\n]+)\nTimes: User: +([0-9.]+)s System: +([0-9.]+)s Elapsed: +([0-9:]+) *")
    term = rterm.search(text)
    result = dict()
    if term:
        result["termination_text"] = term.group(0)
        result["termination_name"] = term.group(1)
        result["termination_message"] = term.group(2)
        result["usertime"] = term.group(3)
        result["systemtime"] = term.group(4)
        result["elapsedtime"] = term.group(5)
        result["nlines"] = term.group(0).count("\n")
    return result

# Match "phaser-style" CCP4 banner
def isccp4banner_phaser(text):
    """Test if text matches a 'phaser-style' CCP4 program banner.

    'Phaser-style' banners look similar to CCP4 banners but
    contain some different information. They are also used by
    the 'pointless' program.
    
    If the match fails then return False; if it succeeds then
    return a dictionary with the following keys:

    name: the name of the program from the banner.
    
    version: the reported program version.
    
    user: the user id that appears in the banner at runtime.
    
    rundate: the date that the program run started at as
    reported in the program banner.
    
    runtime: the time of day that the program run started at as
    reported in the program banner.
    
    os: corresponds to the 'os type' as reported in the banner,
    for example 'linux'.
    
    date: corresponds to the 'release date' of the program as
    reported in the banner.

    ccp4version: currently set to '?'."""
    # This style of banner looks like:
    # 1234567890123456789012345678901234567890123456789012345678901234567890123456789012345
    # #####################################################################################
    # #####################################################################################
    # #####################################################################################
    # ### CCP4 PROGRAM SUITE: Phaser                                              1.3.2 ###
    # #####################################################################################
    # User:         pjx
    # Run time:     Wed May 17 09:27:42 2006
    # Version:      1.3.2
    # OS type:      linux
    # Release Date: Sun Feb  5 17:29:18 2006
    #
    # Note that Pointless also writes a similar banner, but with "CCP4 SUITE"
    # substituted for "CCP4 PROGRAM SUITE"
    rbanner = re.compile(r"#{85,85}\n#{85,85}\n#{85,85}\n### CCP4 PROGRAM SUITE: ([A-Za-z0-9_.]+) *([0-9.]+) *###\n#{85,85}\nUser: *([^ ]+)\nRun time: *([A-Za-z0-9: /]+)\nVersion: *([0-9.]+)\nOS type: *([^\n]*)\nRelease Date: *([A-Za-z0-9: /]+)")
    rbanner = re.compile(r"#+\n#+\n#+\n### CCP4 (PROGRAM )?SUITE: ([A-Za-z0-9_.]+) *([0-9.]+) *###\n#+\nUser: *([^ ]+)\nRun time: *([A-Za-z0-9: /]+)\nVersion: *([0-9.]+)\nOS type: *([^\n]*)\nRelease Date: *([A-Za-z0-9: /]+)")
    banner = rbanner.search(text)
    result = dict()
    if banner:
        ##print "Identified Phaser-style banner"
        result["banner_text"] = banner.group(0)
        result["name"] = banner.group(2)
        result["version"] = banner.group(3)
        result["user"] = banner.group(4)
        result["rundate"] = banner.group(5)
        result["runtime"] = banner.group(5)
        result["os"] = banner.group(7)
        result["date"] = banner.group(8)
        result["ccp4version"] = "?"
        result["nlines"] = banner.group(0).count("\n")
    return result

# Match "phaser-style" CCP4 program termination
def isccp4termination_phaser(text):
    """Test if text matches a 'phaser-style' CCP4 program termination.

    If the match fails then return False; if it succeeds then
    return a dictionary with the following keys:

    termination_name: the program name as reported in the
    CCP4 termination message at the tail of the program log.

    termination_message: the message text displayed in the
    CCP4 termination message, e.g. 'SUCCESS'.

    systemtime: the value of the 'CPU time' given at
    termination.

    Note that this is a subset of the attributes collected
    for standard CCP4 termination messages."""
    # This style of banner looks like:
    # 12345678012345678012
    # --------------------
    # EXIT STATUS: SUCCESS
    # --------------------
    #
    # CPU Time: 0 days 0 hrs 1 mins 34.43 secs (94.43 secs)
    # Finished: Wed May 17 09:29:25 2006
    rterm = re.compile(r"\-*\nEXIT STATUS: *([^\n]+)\n\-*\n\nCPU Time: *([A-Za-z0-9 \.\(\)]*)\nFinished: *([A-Za-z0-9 \.\(\)]*)")
    term = rterm.search(text)
    result = dict()
    if term:
        result["termination_text"] = term.group(0)
        result["termination_message"] = term.group(1)
        result["systemtime"] = term.group(2)
        result["nlines"] = term.group(0).count("\n")
    return result

# Match CCP4 keytext i.e. $TEXT ...
def isccp4keytext(text):
    #
    # See e.g. http://www.ccp4.ac.uk/dist/html/loggraphformat.html
    # for format of TEXT information, but essentially it's:
    #
    # $TEXT :text name: $$ junk (ignored) text $$any text characters$$
    #
    rkeytext = re.compile(r"\$TEXT[ \n]*:([^:]*):[ \n]*\$\$([^\$]*)\$\$([^\$]*)\$\$")
    keytext = rkeytext.search(text)
    result = dict()
    if keytext:
        ##print "Located CCP4 TEXT"
        ##print keytext.group(0)
        result["name"] = keytext.group(1)
        result["junk_text"] = keytext.group(2)
        result["message"] = keytext.group(3)
        result["nlines"] = keytext.group(0).count("\n")
    return result

# Match CCP4 TABLE
def isccp4table(text):
    ##print "************* TEXT **************"
    ##print text
    ##print "*********************************"
    #
    # See e.g. http://www.ccp4.ac.uk/dist/html/loggraphformat.html
    # for format of TABLES
    #
    # Set up regular expression for entire table
    ##rtable = re.compile(r" *\$TABLE ?:([^:]*):[ \n]+\$(GRAPHS|SCATTER)[ \n]?(:[^\$]*)\$\$[ \n]+([^\$]*)\$\$[ \n]+([^\$]*)\$\$[ \n]+([^\$]*)\$\$")
    rtable = re.compile(r" *\$TABLE ?:([^:]*):[ \n]+\$(GRAPHS|SCATTER)[^:]*(:[^\$]*)\$\$([^\$]*)\$\$([^\$]*)\$\$([^\$]*)\$\$")
    ##if rtable1.search(text):
    ##    print "*********************************"
    ##    print "Partial table:"
    ##    print text
    ##    print "*********************************"
    table = rtable.search(text)
    result = dict();
    if table:
        result["rawtable"] = table.group(0)
        result["title"] = table.group(1).strip()
        result["type"] = table.group(2).strip()
        result["graphs"] = table.group(3)
        result["columns"] = table.group(4)
        result["text"] = table.group(5)
        result["data"] = table.group(6)
        result["nlines"] = table.group(0).count("\n")
    return result

# Match CCP4i header information
def isccp4iheader(text):
    #
    # CCP4i header lines look like:
    # #CCP4I VERSION CCP4Interface 1.4.1
    # #CCP4I SCRIPT LOG refmac5
    # #CCP4I DATE 25 Oct 2005  15:19:22
    #
    if isccp4itail(text):
        # Reject tail elements
        return ""
    rheader = re.compile(r"#CCP4I (.*)")
    header = rheader.search(text)
    result = ""
    if header:
        result = header.group(0)
    return result

# Match CCP4i header information
def isccp4itail(text):
    #
    # CCP4i tail lines look like:
    # #CCP4I TERMINATION STATUS 1
    # #CCP4I TERMINATION TIME 25 Oct 2005  15:19:30
    # #CCP4I TERMINATION OUTPUT_FILES  /home/pjx/PROJECTS/myProject/...
    # #CCP4I MESSAGE Task completed successfully
    #
    rtail = re.compile(r"#CCP4I (TERMINATION|MESSAGE) (.*)")
    tail = rtail.search(text)
    result = ""
    if tail:
        result = tail.group(0)
    return result

# Match CCP4i information text
def isccp4i_information(text):
    #
    # CCP4i information lines look like:
    # 123456789012345678901234567890123456789012345678901234567890123456789012345
    # ***************************************************************************
    # * Information from CCP4Interface script
    # ***************************************************************************
    # Running SHELXC to prepare data for heavy atom search
    # ***************************************************************************
    #
    ##print "************* TEXT **************"
    ##print text
    ##print "*********************************"
    rinfo = re.compile(r"\*{75,75}\n\* Information from CCP4Interface script\n\*{75,75}\n(.*)\n\*{75,75}")
    info = rinfo.search(text)
    result = dict()
    if info:
        result["message"] = info.group(1)
        result["nlines"] = info.group(0).count("\n")
    return result

# Match SHELX banners
#  123456789012345678901234567890123456789012345678901234567890123456789012
#  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  +  SHELXC - Create input files for SHELXD and SHELXE - Version 2006/3  +
#  +  Copyright (C) George M. Sheldrick 2003-6                            +
#  +  SHELX_56_shelxc                 Started at 14:30:07 on 21 Apr 2006  +
#  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#
#  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  +  SHELXD-2006/3 - MACROMOLECULAR DIRECT METHODS - FORTRAN-95 VERSION  +
#  +  Copyright (C)  George M. Sheldrick 2000-2006                        +
#  +  SHELX_56_shelxd_fa              started at 14:30:11 on 21 Apr 2006  +
#  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#
#  12345678901234567890123456789012345678901234567890123456789012345678
#  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  +  SHELXE  -  PHASING AND DENSITY MODIFICATION  -  Version 2006/3  +
#  +  Copyright (C)  George M. Sheldrick 2001-6                       +
#  +  Started at 14:30:36 on 21 Apr 2006                              +
#  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#
def isshelxbanner(text):
    """Test if text matches a SHELX program banner.

    This function tries to match the banners from SHELXC,
    SHELXD and SHELXE.

    If the match fails then return False; if it succeeds then
    return a dictionary with the following keys:

    name: the program name.

    version: the program version.

    runtime: the time of day that the program run started at as
    reported in the program banner.

    rundate: the date that the program run started at as
    reported in the program banner."""

    # Set up regular expression for partial SHELX banner
    rbanner = re.compile(r"\+{68,72}\n  \+  (SHELXC|SHELXD|SHELXE)([^\+]*)\+\n  \+  Copyright \(C\) *George M. Sheldrick[^\n]*\n  \+  ([^\+]*)\+\n  \+{68,72}")
    banner = rbanner.search(text)
    result = dict()
    if banner:
        #print "************* TEXT **************"
        #print text
        #print "*********************************"
        result["banner_text"] = banner.group(0)
        result["name"] = banner.group(1)
        result["nlines"] = banner.group(0).count("\n")
        # Try and untangle the versions
        name = banner.group(1)
        version = False
        if name == "SHELXC" or name == "SHELXE":
            version = re.compile(r"Version ([^ \+]*)").search(banner.group(2))
        elif name == "SHELXD":
            version = re.compile(r"\-([^ \-]*)").search(banner.group(2))
        if version:
            result["version"] = version.group(1)
        else:
            result["version"] = "?"
        # Try and untangle the start times
        date = re.compile(r"(S|s)tarted at ([0-9:]+) on ([0-9A-Za-z ]+)").search( \
            banner.group(3))
        if date:
            result["runtime"] = str(date.group(2))
            result["rundate"] = str(date.group(3))
        else:
            result["runtime"] = "?"
            result["rundate"] = "?"
    return result

# Match SHELX program termination
def isshelxtermination(text):
    """Test if text matches a SHELX program termination.

    This function tries to match the messages from SHELXC,
    SHELXD and SHELXE.

    If the match fails then return False; if it succeeds then
    return a dictionary with the following keys:

    termination_name: the program name as reported in the
    SHELX termination message at the tail of the program log.
    
    termination_message: the message text displayed in the
    termination message. The content of this text varies
    between SHELXC and SHELXD/E so no further processing is
    currently attempted."""
    #print "************* TEXT **************"
    #print text
    #print "*********************************"
    #
    # Termination looks like:
    #  123456789012345678901234567890123456789012345678901234567890123456789012
    #  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    #  +  SHELXC for SHELX_56_shelxc finished at 14:30:11 on 21 Apr 2006      +
    #  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    #
    #  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    #  +  SHELXD finished at 14:30:36      Total time:        23.43 secs  +
    #  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    #
    #  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    #  +  SHELXE finished at 14:34:01      Total time:       198.15 secs  +
    #  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    #
    # Set up regular expression for partial SHELX termination
    rterm = re.compile(r"\+{68,72}\n  \+  (SHELXC|SHELXD|SHELXE)([^\+]*)\+\n  \+{68,72}")
    term = rterm.search(text)
    result = dict()
    if term:
        result["termination_text"] = term.group(0)
        result["termination_name"] = term.group(1)
        result["termination_message"] = term.group(2)
        result["nlines"] = term.group(0).count("\n")
    return result

#######################################################################
# External Functions
#######################################################################

# parselog
#
# Given the name of a logfile, populates and returns a
# logfile object

def parselog(filen,progress=0):
    """Process a file and return a populated logfile object.

    parselog takes a file name as input; optionally if the
    progress argument is set to a positive integer then the
    function also reports its progress when it reaches a
    multiple of that number of lines.

    parselog works by reading the source file one line at a
    time from beginning to end. Each line is added to two
    buffers: a 'small' buffer, which stores the last 10 lines
    read, and a 'large' tablebuffer, which can store the last
    1000 lines.

    After a line has been added, the small buffer is checked
    against a series of regular expressions designed to match
    the various features (banners, terminations and so on).
    If a match is found then parselog updates the logfile object
    that it is constructing and then clears the buffer.
    The tablebuffer is also checked at each line, to see if it
    contains a whole CCP4 logfile table. (The tablebuffer is a
    specialised form of buffer which is intended to optimise
    dealing with tables).

    The buffer sizes for the small and large buffers affect the
    speed of operation of parselog - if they are large then the
    parsing is slower because larger chunks of text are being
    tested multiple times.
    However if the buffers are too small to accommodate some
    of the logfile features then parselog is unable to detect
    those features. As some logfiles can contain extremely large
    tables, the tablebuffer must also be large. However other
    features are generally quite small.

    The other factor that can affect the speed of parsing is the
    make-up of the logfile. Counterintuitively, long files that
    contain few recognisable features can take longer because the
    buffers are only infrequently flushed."""
    
    # Process a file and return a populated logfile object
    #
    # Maximum size of text buffers to use
    bufsize = 10
    tablebufsize = 1000
    # Initial size of chunks to process 
    chunksize = 10
    # Buffer objects
    buff = buffer(bufsize)
    tablebuff = tablebuffer(tablebufsize)
    linecount = 0
    # New (empty) logfile object
    log = logfile(filen)
    prog = False
    # Open the file for reading
    f = open(filen,"r")
    # Read line-by-line
    for line in f:
        linecount += 1
        # Progress indicator (if requested)
        # Report reaching "progress" number of lines
        if progress:
            if not linecount % progress:
                print "Processed "+str(linecount)+" lines"
        # Append line to buffers
        buff.append(line)
        tablebuff.append(line)
        # Get a chunk of text to process
        bufftext = buff.tail(chunksize)
        # Test the buffer for matches
        #
        # CCP4 program banner
        result = isccp4banner(bufftext)
        if result:
            ##print "Found CCP4 program banner"
            ##print "Result = "+str(result)
            prog = log.addprogram()
            prog.set_isccp4(True)
            prog.set_attributes_from_dictionary(result)
            log.set_fragment_start(linecount)
            buff.clear()
            tablebuff.clear()
            continue
        # SHELX program banner
        result = isshelxbanner(bufftext)
        if result:
            ##print "Found SHELX program banner"
            ##print "Result = "+str(result)
            prog = log.addprogram()
            prog.set_attributes_from_dictionary(result)
            log.set_fragment_start(linecount)
            buff.clear()
            tablebuff.clear()
            continue
        # CCP4 program termination
        result = isccp4termination(bufftext)
        if result:
            ##print "Found CCP4 program termination"
            ##print "Result = "+str(result)
            if not prog:
                # Found the end of a program before its start
                prog = log.addprogram()
            prog.set_attributes_from_dictionary(result)
            log.set_fragment_end(linecount)
            prog.set_termination(True)
            # Clear the current pointer
            prog = False
            buff.clear()
            tablebuff.clear()
            continue
        # SHELX program termination
        result = isshelxtermination(bufftext)
        if result:
            ##print "Found SHELX program termination"
            ##print "Result = "+str(result)
            if not prog:
                # Found the end of a program before its start
                prog = log.addprogram()
            prog.set_attributes_from_dictionary(result)
            log.set_fragment_end(linecount)
            prog.set_termination(True)
            # Clear the current pointer
            prog = False
            buff.clear()
            tablebuff.clear()
            continue
        # CCP4 table
        if tablebuff.complete():
            tablebufftext = tablebuff.tail(tablebufsize)
            result = isccp4table(tablebufftext)
            if result:
                ##print "Found CCP4 table"
                ##print "Result = "+str(result)
                if not prog:
                    # Found a table outside the context of a program
                    ##print "Adding table as a fragment"
                    prog = log.newfragment()
                    log.set_fragment_start(linecount)
                table = prog.addtable(result["rawtable"])
                log.addtable(table)
                # clear the buffers
                buff.clear()
                tablebuff.clear()
                continue
        # CCP4 keytext message
        result = isccp4keytext(bufftext)
        if result:
            ##print "Found CCP4 keytext"
            ##print "Result = "+str(result)
            if not prog:
                # Found a message outside the context of a program
                ##print "Adding keytext as a fragment"
                prog = log.addfragment()
                log.set_fragment_start(linecount)
            keytext = prog.addkeytext(result["name"], \
                                      result["junk_text"], \
                                      result["message"])
            log.addkeytext(keytext)
            buff.clear()
            tablebuff.clear()
            continue
        # CCP4i header
        result = isccp4iheader(bufftext)
        if result:
            ##print "Found CCP4i header"
            ##print "Result = "+str(result)
            log.append_ccp4i_header(result)
            buff.clear()
            continue
        # CCP4i tail
        result = isccp4itail(bufftext)
        if result:
            ##print "Found CCP4i tail"
            ##print "Result = "+str(result)
            log.append_ccp4i_tail(result)
            buff.clear()
            tablebuff.clear()
            continue
        # CCP4i information
        result = isccp4i_information(bufftext)
        if result:
            ##print "Found CCP4i information"
            ##print "Result = "+str(result)
            # Make a new fragment - these messages shouldn't
            # appear inside the context of another program
            prog = log.addccp4i_info()
            prog.set_attributes_from_dictionary(result)
            log.set_fragment_start(linecount)
            log.set_fragment_end(linecount)
            # Clear the current context
            prog = False
            buff.clear()
            tablebuff.clear()
            continue
    # Ensure that the endline of the last fragment
    # is assigned
    log.set_fragment_end(linecount)
    # Close the file
    f.close()
    return log

#
# summarise
#
# Produce a summary of the data in a logfile object
#
def summarise(thislog):
    """Summarise the content of a logfile object.

    This function takes a logfile object as input and writes a
    a summary of the contents (fragments, programs, tables, messages
    and so on) to stdout."""

    # Logfile name
    print "Summary for "+thislog.filename()+"\n"
    # Was it from CCP4i?
    if thislog.isccp4i():
        print "This is a CCP4i logfile\n"
    # Number of programs or pseudo-programs
    print str(thislog.nfragments())+" logfile fragments\n"
    print "Fragments:"
    for i in range(0,thislog.nfragments()):
        fragment = thislog.fragment(i)
        if fragment.isprogram():
            print "\tProgram: "+str(fragment.name)
        else:
            if fragment.isccp4i_info():
                print "\tCCP4i info"
            elif fragment.isfragment():
                print "\tFragment"
            if fragment.ntables():
                print "\t\t"+str(fragment.ntables())+" tables"
            if fragment.nkeytexts():
                print "\t\t"+str(fragment.nkeytexts())+" keytexts"
                    
    print ""
    # Summarise program logfile fragments
    if thislog.nprograms() > 0:
        print str(thislog.nprograms())+" program logfiles\n"
        print "Programs:"
        for i in range(0,thislog.nprograms()):
            prog = thislog.program(i)
            # Is it a CCP4 program?
            if prog.isccp4():
                # Print name, version (and CCP4 version)
                print "\t"+prog.name+ \
                      "\tv"+prog.version+ \
                      "\t(CCP4 "+prog.ccp4version+")"
            else:
                # Print name and version
                print "\t"+prog.name+"\t"+prog.version
            if prog.termination():
                print "\tTerminated with: "+prog.termination_message
            else:
                print "\tNo termination message found"
            # Keytexts
            if prog.nkeytexts():
                print "\n\t\tKeytext messages:"
                for j in range(0,prog.nkeytexts()):
                    print "\t\t"+str(prog.keytext(j).name())+ \
                          ": \""+str(prog.keytext(j).message())+"\""
            # Tables
            if prog.ntables():
                print "\n\t\tTables:"
                for j in range(0,prog.ntables()):
                    print "\t\tTable: \""+prog.table(j).title()+"\""
            print ""
    else:
        print "No program logfiles found"
    print ""
    # Total set of CCP4i information messages in the file
    print "CCP4i messages in file:"
    if thislog.nccp4i_info():
        for i in range(0,thislog.nccp4i_info()):
            print "\tCCP4i info: \""+thislog.ccp4i_info(i).message+"\""
    else:
        print "\tNo messages found"
    print ""
    # Total set of tables in the file
    print "Tables in file:"
    if thislog.ntables():
        for i in range(0,thislog.ntables()):
            print "\tTable: \""+thislog.table(i).title()+"\""
    else:
        print "\tNo tables found"
    print ""
    # Total set of keytexts in the file
    print "Keytext messages in file:"
    if thislog.nkeytexts():
        for i in range(0,thislog.nkeytexts()):
            print "\t"+str(thislog.keytext(i).name())+ \
                  ": \""+thislog.keytext(i).message()+"\""
    else:
        print "\tNo keytext messages found"
    print ""
            
    
##############################################################
# Diagnostic methods used for testing and as examples
##############################################################

# List the TABLE tags in a file
def table_tags(filen):
    """Report occurances of '$TABLE' tags in a log file.

    This function is principally a diagnostic tool and is
    independent of the other classes and methods in this
    module. It takes the name of a log file as input,
    scans the file for occurances of the $TABLE tag, and
    reports this to stdout."""
    
    print "Scanning file "+str(filen)
    rtable = re.compile(r"\$TABLE *:")
    f = open(filen,"r")
    linecount = 0
    tablecount = 0
    tablelist = []
    for line in f:
        linecount = linecount + 1
        table = rtable.search(line)
        if table:
            tablecount = tablecount + 1
            print str(linecount)+": "+str(line.rstrip("\n"))
            tablelist.append(line.rstrip("\n"))
    f.close()
    print str(linecount)+" lines and "+str(tablecount)+" tables"
    return tablelist

# An example of making a new table from scratch
def table_example():
    """Demonstration function that creates and populates a table object.

    This function is for demonstration purposes only; it shows
    the basics of how to make and output a table. It creates
    a new table object, names it, populates some columns of data
    and then adds some graph definitions before outputting the
    formatted table to stdout."""
    
    print "\nExample making a new table from scratch:\n"
    tbl = table()
    tbl.settitle("A table with random data")
    tbl.settype("GRAPH")
    for i in range(0,3):
        col = tbl.addcolumn("col_"+str(i))
        for j in range(0,6):
            col.append(i*j)
    g1 = tbl.addgraph("An arbitrary graph")
    g1.setscaling("N")
    g1.setcolumns("1,2")
    g2 = tbl.addgraph("Another arbitrary graph")
    g2.setscaling("A")
    g2.setcolumns("1,3")
    print tbl.show()

def report_example(log):
    print "\n****** Report on "+filen+" ******\n"
    print "Number of programs = "+str(log.nprograms())
    for i in range(0,log.nprograms()):
        if log.program(i):
            # Program name, program version and CCP4 version
            print "\t"+log.program(i).name+ \
                  "\t"+str(log.program(i).version)+ \
                  "\t"+str(log.program(i).ccp4version)
            # Tables and graphs
            if log.program(i).ntables() > 0:
                print "\t"+str(log.program(i).ntables())+" tables:"
                for j in range(0,log.program(i).ntables()):
                    # Print each table
                    print "\t\t"+log.program(i).table(j).title()
                    print "\t\t"+str(log.program(i).table(j).ncolumns()) \
                          +" columns:"
                    for k in range(0,log.program(i).table(j).ncolumns()):
                        # List the columns
                        print "\t\t\t" \
                              +log.program(i).table(j).table_column(k).title() \
                              +" " \
                              +str(log.program(i).table(j).table_column(k).data())
                    print "\t\t"+str(log.program(i).table(j).ngraphs()) \
                          +" graphs:"
                    for k in range(0,log.program(i).table(j).ngraphs()):
                        # List the graphs
                        print "\t\t\t" \
                              +log.program(i).table(j).table_graph(k).title() \
                              +" " \
                              +str(log.program(i).table(j).table_graph(k).columns())
                    # Dump the table from the object
                    # using the "show" method
                    print log.program(i).table(j).show()
        else:
            print "\t*** Null program at position "+str(i)
            print ""


if __name__ == "__main__":
    """Usage example and demonstration for smartie.

    Run the main program using:
    
    python smartie.py file1 [file2 [...] ]

    For each file this example will generate a logfile object
    and then use the logfile 'summarise' method to print out
    a summary of the file's contents."""
    
    print "Running test on logparser code"
    # Get the command line args
    print "command line: "+str(sys.argv)
    if len(sys.argv) == 1:
        print "Usage: smartie.py file1 [file2 [...] ]"
        sys.exit(0)
    # Cycle over files and process
    for filen in sys.argv[1:]:
        print "**** Parsing file \""+filen+"\""
        log = parselog(filen)
        # Use the summarise function
        summarise(log)
        
